#!/usr/bin/env python3

import fcntl
import os
import pathlib
import pty
import shutil
import subprocess
import sys
import termios
import time


CRIU_BIN = "../../../criu/criu"
CRIU_NS = "../../../scripts/criu-ns"
IMG_DIR = "dumpdir"
DUMP_LOG = "dump.log"
RESTORE_LOG = "restore.log"
PIDFILE = "pidfile"


def check_dumpdir(path=IMG_DIR):
    if os.path.isdir(path):
        shutil.rmtree(path)
    os.mkdir(path, 0o755)


def run_task_with_own_pty(task):
    fd_m, fd_s = pty.openpty()

    pid = os.fork()
    if pid == 0:
        os.close(fd_m)
        os.setsid()
        os.dup2(fd_s, 0)
        os.dup2(fd_s, 1)
        os.dup2(fd_s, 2)
        fcntl.ioctl(fd_s, termios.TIOCSCTTY, 1)
        os.close(fd_s)
        task()
        exit(0)

    os.close(fd_s)
    fd_m = os.fdopen(fd_m, "rb")
    os.set_blocking(fd_m.fileno(), False)

    while True:
        try:
            data = fd_m.read()
        except IOError:
            break
        if data is not None:
            print(data.decode("utf-8"))

    _, status = os.waitpid(pid, 0)

    try:
        data = fd_m.read()
    except IOError as err:
        print(err)

    if data is not None:
        print(data.decode("utf-8"))
    fd_m.close()

    if status != 0:
        print("task %s exited badly: %d" % (task.__name__, status))
        exit(1)

    return 0


def create_pty():
    fd_m, fd_s = pty.openpty()
    return (os.fdopen(fd_m, "wb"), os.fdopen(fd_s, "wb"))


def create_isolated_dumpee():
    pathlib.Path("running").touch()
    fd_m, fd_s = create_pty()
    pid = os.fork()
    if pid == 0:
        os.setsid()
        os.dup2(fd_s.fileno(), 0)
        os.dup2(fd_s.fileno(), 1)
        os.dup2(fd_s.fileno(), 2)
        fcntl.ioctl(fd_s.fileno(), termios.TIOCSCTTY, 1)
        while True:
            if not os.access("running", os.F_OK):
                sys.exit(0)
            time.sleep(1)
    fd_m.close()
    fd_s.close()
    return pid


def criu_ns_dump(pid, shell_job=False):
    cmd = [CRIU_NS, "dump", "-D", IMG_DIR, "-v4", "-t", str(pid),
           "--log-file", DUMP_LOG, "--criu-binary", CRIU_BIN]
    if shell_job:
        cmd.append("--shell-job")
    ret = subprocess.Popen(cmd).wait()
    return ret


def criu_ns_restore(shell_job=False, restore_detached=False):
    cmd = [CRIU_NS, "restore", "-D", IMG_DIR, "-v4", "--log-file",
           RESTORE_LOG, "--criu-binary", CRIU_BIN]
    if shell_job:
        cmd.append("--shell-job")
    if restore_detached:
        cmd += ["--restore-detached", "--pidfile", PIDFILE]
    ret = subprocess.Popen(cmd).wait()
    return ret


def read_log_file(filename):
    logfile_path = os.path.join(IMG_DIR, filename)
    with open(logfile_path) as logfile:
        print(logfile.read())


def test_dump_and_restore_with_shell_job():
    print("Test criu-ns dump and restore with --shell-job option")
    check_dumpdir()
    pathlib.Path("running").touch()
    pid = os.fork()
    if pid == 0:
        while True:
            if not os.access("running", os.F_OK):
                sys.exit(0)
            time.sleep(1)

    ret = criu_ns_dump(pid, shell_job=True)
    if ret != 0:
        read_log_file(DUMP_LOG)
        sys.exit(ret)

    os.unlink("running")
    fd_m, fd_s = create_pty()
    pid = os.fork()
    if pid == 0:
        os.setsid()
        fd_m.close()
        # since criu-ns takes control of the tty stdin
        os.dup2(fd_s.fileno(), 0)
        ret = criu_ns_restore(shell_job=True)
        if ret != 0:
            read_log_file(RESTORE_LOG)
            sys.exit(ret)
        os._exit(0)

    fd_s.close()
    os.waitpid(pid, 0)


def test_dump_and_restore_without_shell_job(restore_detached=False):
    print("Test criu-ns dump and restore with an isolated process"
          "(%d)" % restore_detached)
    check_dumpdir()
    pid = create_isolated_dumpee()
    ret = criu_ns_dump(pid)
    if ret != 0:
        read_log_file(DUMP_LOG)
        sys.exit(ret)

    if not restore_detached:
        os.unlink("running")

    pid = os.fork()
    if pid == 0:
        os.setsid()
        ret = criu_ns_restore(restore_detached=restore_detached)
        if ret != 0:
            read_log_file(RESTORE_LOG)
            sys.exit(ret)
        os._exit(0)

    os.waitpid(pid, 0)


def test_dump_and_restore_in_pidns():
    if os.system("grep NSpid /proc/self/status"):
        return

    print("Test criu-ns dump and restore in namespaces")

    def _dump():
        pid = create_isolated_dumpee()
        ret = criu_ns_dump(pid)
        if ret != 0:
            read_log_file(DUMP_LOG)
            sys.exit(ret)

    def _restore():
        ret = criu_ns_restore(restore_detached=True)
        if ret != 0:
            read_log_file(RESTORE_LOG)
            sys.exit(ret)

    def _get_restored_pid():
        restored_pid = 0
        pidfile_path = os.path.join(IMG_DIR, PIDFILE)
        if not os.path.exists(pidfile_path):
            raise FileNotFoundError("pidfile not found")
        with open(pidfile_path, "r") as pidfile:
            restored_pid = pidfile.read().strip()
        return int(restored_pid)

    def _redump():
        global IMG_DIR
        try:
            restored_pid = _get_restored_pid()
        except FileNotFoundError:
            sys.exit(1)
        IMG_DIR = "dumpdir2"
        check_dumpdir(IMG_DIR)
        ret = criu_ns_dump(restored_pid)
        if ret != 0:
            read_log_file(DUMP_LOG)
            sys.exit(ret)

    def _re_restore():
        os.unlink("running")
        ret = criu_ns_restore()
        if ret != 0:
            read_log_file(RESTORE_LOG)
            sys.exit(ret)

    check_dumpdir()
    _dump()
    _restore()
    _redump()
    _re_restore()


def main():
    test_dump_and_restore_with_shell_job()
    test_dump_and_restore_without_shell_job()
    test_dump_and_restore_without_shell_job(restore_detached=True)
    test_dump_and_restore_in_pidns()


if __name__ == "__main__":
    run_task_with_own_pty(main)



#!/usr/bin/python3
"""
`pip uninstall` doesn't support `--prefix`.
https://github.com/pypa/pip/issues/11213
"""
import argparse
import os
import shutil
import site
import subprocess
import sys

import importlib_metadata


def add_site_dir(prefix: str):
    """
    Add site directory with prefix to sys.path and update PYTHONPATH.
    """
    # If prefix is used, we need to make sure that we
    # do not uninstall other packages from the system paths.
    sys.path = []
    site.PREFIXES = [prefix]
    pkgs = site.getsitepackages()
    for path in pkgs:
        site.addsitedir(path)
        if 'dist-packages' in path:
            # Ubuntu / Debian might use both dist- and site- packages.
            site.addsitedir(path.replace('dist-packages', 'site-packages'))
    os.environ['PYTHONPATH'] = os.pathsep.join(sys.path)


def uninstall_module(package_name: str, prefix=None):
    """
    Enable support for '--prefix' with 'pip uninstall'.
    """
    dist_info_path = None
    if prefix:
        add_site_dir(prefix)
        try:
            distribution = next(importlib_metadata.Distribution.discover(name=package_name))
            dist_info_path = str(distribution._path)
        except StopIteration:
            print(f"Skipping {package_name} as it is not installed.")
            sys.exit(0)

    command = [sys.executable, '-m', 'pip', 'uninstall', '-y', package_name]
    try:
        subprocess.check_call(command, env=os.environ)
        if dist_info_path and os.path.isdir(dist_info_path):
            # .dist-info files are not cleaned up when the package
            # has been installed with --prefix.
            # https://github.com/pypa/pip/issues/5573
            shutil.rmtree(dist_info_path)
            if 'dist-packages' in dist_info_path:
                shutil.rmtree(dist_info_path.replace('dist-packages', 'site-packages'))
    except subprocess.CalledProcessError as err:
        print(f'Error uninstalling package {package_name}: {err}')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('module_name', help='The name of the module to uninstall')
    parser.add_argument('--prefix', help='The prefix where the module was installed')
    args = parser.parse_args()
    uninstall_module(args.module_name, args.prefix)



#!/usr/bin/python3
import sys
import re

re_file = r'^diff --git a/(\S\S*)\s.*$'
re_line = r'^@@ -(\d\d*)\D.*@@.*$'

if __name__ == '__main__':
    if len(sys.argv) != 1 and len(sys.argv) != 2:
        print(f'usage: {sys.argv[0]} <path/to/file>')
        print(f'usage: <command> | {sys.argv[0]}')
        exit(1)

    input_file = sys.stdin.fileno()
    if len(sys.argv) == 2:
        input_file = sys.argv[1]

    with open(input_file, 'r') as fi:
        file_name = None
        line_number = None
        for line in fi:
            file_matches = re.findall(re_file, line)
            if len(file_matches) == 1:
                file_name = file_matches[0]
                continue

            if file_name is None:
                continue

            line_matches = re.findall(re_line, line)
            if len(line_matches) == 1:
                line_number = int(line_matches[0]) + 3
                print(f'::warning file={file_name},line={line_number}::clang-format: Possible coding style problem (https://github.com/checkpoint-restore/criu/blob/criu-dev/CONTRIBUTING.md#automatic-tools-to-fix-coding-style)')



import os
import tempfile
import subprocess


class criu_config:
    @staticmethod
    def run(action,
            args,
            criu_bin,
            fault=None,
            strace=[],
            preexec=None,
            preload=False,
            nowait=False):

        config_path = tempfile.mktemp(".conf", "criu-%s-" % action)
        with open(config_path, "w") as config_fd:
            for arg in args:
                if arg.startswith("--"):
                    config_fd.write("\n")
                    arg = arg.strip("-")
                config_fd.write("%s " % arg)

        env = dict(
            os.environ,
            ASAN_OPTIONS="log_path=asan.log:disable_coredump=0:detect_leaks=0"
        )

        if fault:
            print("Forcing %s fault" % fault)
            env['CRIU_FAULT'] = fault

        cr = subprocess.Popen(
            strace +
            [criu_bin, action, "--no-default-config", "--config", config_path],
            env=env,
            close_fds=False,
            preexec_fn=preexec
        )
        if nowait:
            return cr
        return cr.wait()

    @staticmethod
    def exit_signal(ret):
        return ret < 0



# Define structures and constants for generating elf file.
import ctypes

Elf64_Half = ctypes.c_uint16  # typedef uint16_t Elf64_Half;
Elf64_Word = ctypes.c_uint32  # typedef uint32_t Elf64_Word;
Elf64_Addr = ctypes.c_uint64  # typedef uint64_t Elf64_Addr;
Elf64_Off = ctypes.c_uint64  # typedef uint64_t Elf64_Off;
Elf64_Xword = ctypes.c_uint64  # typedef uint64_t Elf64_Xword;

# Elf64_Ehdr related constants.

# e_ident size.
EI_NIDENT = 16  # #define EI_NIDENT (16)

EI_MAG0 = 0  # #define EI_MAG0         0               /* File identification byte 0 index */
ELFMAG0 = 0x7f  # #define ELFMAG0         0x7f            /* Magic number byte 0 */

EI_MAG1 = 1  # #define EI_MAG1         1               /* File identification byte 1 index */
ELFMAG1 = ord('E')  # #define ELFMAG1  'E'             /* Magic number byte 1 */

EI_MAG2 = 2  # #define EI_MAG2         2               /* File identification byte 2 index */
ELFMAG2 = ord('L')  # #define ELFMAG2  'L'             /* Magic number byte 2 */

EI_MAG3 = 3  # #define EI_MAG3         3               /* File identification byte 3 index */
ELFMAG3 = ord('F')  # #define ELFMAG3  'F'             /* Magic number byte 3 */

EI_CLASS = 4  # #define EI_CLASS        4               /* File class byte index */

EI_DATA = 5  # #define EI_DATA         5               /* Data encoding byte index */

EI_VERSION = 6  # #define EI_VERSION      6               /* File version byte index */

ELFDATA2LSB = 1  # #define ELFDATA2LSB     1               /* 2's complement, little endian */

ELFCLASS64 = 2  # #define ELFCLASS64      2               /* 64-bit objects */

# Legal values for e_type (object file type).
ET_CORE = 4  # #define ET_CORE         4               /* Core file */

# Legal values for e_machine (architecture).
EM_X86_64 = 62  # #define EM_X86_64       62              /* AMD x86-64 architecture */

# Legal values for e_version (version).
EV_CURRENT = 1  # #define EV_CURRENT      1               /* Current version */


class Elf64_Ehdr(ctypes.Structure):  # typedef struct
    _fields_ = [
        ("e_ident",
         ctypes.c_ubyte * EI_NIDENT),  # unsigned char e_ident[EI_NIDENT];
        ("e_type", Elf64_Half),  # Elf64_Half e_type;
        ("e_machine", Elf64_Half),  # Elf64_Half e_machine;
        ("e_version", Elf64_Word),  # Elf64_Word e_version;
        ("e_entry", Elf64_Addr),  # Elf64_Addr e_entry;
        ("e_phoff", Elf64_Off),  # Elf64_Off e_phoff;
        ("e_shoff", Elf64_Off),  # Elf64_Off e_shoff;
        ("e_flags", Elf64_Word),  # Elf64_Word e_flags;
        ("e_ehsize", Elf64_Half),  # Elf64_Half e_ehsize;
        ("e_phentsize", Elf64_Half),  # Elf64_Half e_phentsize;
        ("e_phnum", Elf64_Half),  # Elf64_Half e_phnum;
        ("e_shentsize", Elf64_Half),  # Elf64_Half e_shentsize;
        ("e_shnum", Elf64_Half),  # Elf64_Half e_shnum;
        ("e_shstrndx", Elf64_Half)  # Elf64_Half e_shstrndx;
    ]  # } Elf64_Ehdr;


# Elf64_Phdr related constants.

# Legal values for p_type (segment type).
PT_LOAD = 1  # #define PT_LOAD         1               /* Loadable program segment */
PT_NOTE = 4  # #define PT_NOTE         4               /* Auxiliary information */

# Legal values for p_flags (segment flags).
PF_X = 1  # #define PF_X            (1 << 0)        /* Segment is executable */
PF_W = 1 << 1  # #define PF_W            (1 << 1)        /* Segment is writable */
PF_R = 1 << 2  # #define PF_R            (1 << 2)        /* Segment is readable */


class Elf64_Phdr(ctypes.Structure):  # typedef struct
    _fields_ = [
        ("p_type", Elf64_Word),  # Elf64_Word p_type;
        ("p_flags", Elf64_Word),  # Elf64_Word p_flags;
        ("p_offset", Elf64_Off),  # Elf64_Off p_offset;
        ("p_vaddr", Elf64_Addr),  # Elf64_Addr p_vaddr;
        ("p_paddr", Elf64_Addr),  # Elf64_Addr p_paddr;
        ("p_filesz", Elf64_Xword),  # Elf64_Xword p_filesz;
        ("p_memsz", Elf64_Xword),  # Elf64_Xword p_memsz;
        ("p_align", Elf64_Xword),  # Elf64_Xword p_align;
    ]  # } Elf64_Phdr;


# Elf64_auxv_t related constants.


class _Elf64_auxv_t_U(ctypes.Union):
    _fields_ = [("a_val", ctypes.c_uint64)]


class Elf64_auxv_t(ctypes.Structure):  # typedef struct
    _fields_ = [
        ("a_type",
         ctypes.c_uint64),  # uint64_t a_type;              /* Entry type */
        ("a_un", _Elf64_auxv_t_U)  # union

        # uint64_t a_val;           /* Integer value */
        # /* We use to have pointer elements added here.  We cannot do that,
        # though, since it does not work when using 32-bit definitions
        # on 64-bit platforms and vice versa.  */
        # } a_un;
    ]  # } Elf64_auxv_t;


# Elf64_Nhdr related constants.

NT_PRSTATUS = 1  # #define NT_PRSTATUS  1  /* Contains copy of prstatus struct */
NT_FPREGSET = 2  # #define NT_FPREGSET  2  /* Contains copy of fpregset struct */
NT_PRPSINFO = 3  # #define NT_PRPSINFO  3  /* Contains copy of prpsinfo struct */
NT_AUXV = 6  # #define NT_AUXV  6  /* Contains copy of auxv array */
NT_SIGINFO = 0x53494749  # #define NT_SIGINFO  0x53494749  /* Contains copy of siginfo_t, size might increase */
NT_FILE = 0x46494c45  # #define NT_FILE  0x46494c45  /* Contains information about mapped files */
NT_X86_XSTATE = 0x202  # #define NT_X86_XSTATE  0x202  /* x86 extended state using xsave */


class Elf64_Nhdr(ctypes.Structure):  # typedef struct
    _fields_ = [
        (
            "n_namesz", Elf64_Word
        ),  # Elf64_Word n_namesz;  /* Length of the note's name.  */
        (
            "n_descsz", Elf64_Word
        ),  # Elf64_Word n_descsz;  /* Length of the note's descriptor.  */
        ("n_type", Elf64_Word
         ),  # Elf64_Word n_type;  /* Type of the note.  */
    ]  # } Elf64_Nhdr;


# Elf64_Shdr related constants.


class Elf64_Shdr(ctypes.Structure):
    _fields_ = [
        (
            # Section name (string tbl index)
            "sh_name", Elf64_Word
        ),
        (
            # Section type
            "sh_type", Elf64_Word
        ),
        (
            # Section flags
            "sh_flags", Elf64_Xword
        ),
        (
            # Section virtual addr at execution
            "sh_addr", Elf64_Addr
        ),
        (
            # Section file offset
            "sh_offset", Elf64_Off
        ),
        (
            # Section size in bytes
            "sh_size", Elf64_Xword
        ),
        (
            # Link to another section
            "sh_link", Elf64_Word
        ),
        (
            # Additional section information
            "sh_info", Elf64_Word
        ),
        (
            # Section alignment
            "sh_addralign", Elf64_Xword
        ),
        (
            # Entry size if section holds table
            "sh_entsize", Elf64_Xword
        )
    ]


# elf_prstatus related constants.


# Signal info.
class elf_siginfo(ctypes.Structure):  # struct elf_siginfo
    _fields_ = [
        (
            # Signal number
            "si_signo", ctypes.c_int
        ),
        (
            # Extra code
            "si_code", ctypes.c_int
        ),
        (
            # Errno
            "si_errno", ctypes.c_int
        )
    ]


# A time value that is accurate to the nearest
# microsecond but also has a range of years.
class timeval(ctypes.Structure):  # struct timeval
    _fields_ = [
        (
            # __time_t tv_sec;            /* Seconds.  */
            "tv_sec", ctypes.c_long
        ),
        (
            # __suseconds_t tv_usec;      /* Microseconds.  */
            "tv_usec", ctypes.c_long
        )
    ]


class user_regs_struct(ctypes.Structure):  # struct user_regs_struct
    _fields_ = [
        ("r15",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int r15;
        ("r14",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int r14;
        ("r13",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int r13;
        ("r12",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int r12;
        ("rbp",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int rbp;
        ("rbx",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int rbx;
        ("r11",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int r11;
        ("r10",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int r10;
        ("r9",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int r9;
        ("r8",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int r8;
        ("rax",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int rax;
        ("rcx",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int rcx;
        ("rdx",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int rdx;
        ("rsi",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int rsi;
        ("rdi",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int rdi;
        ("orig_rax", ctypes.c_ulonglong
         ),  # __extension__ unsigned long long int orig_rax;
        ("rip",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int rip;
        ("cs",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int cs;
        ("eflags",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int eflags;
        ("rsp",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int rsp;
        ("ss",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int ss;
        ("fs_base", ctypes.c_ulonglong
         ),  # __extension__ unsigned long long int fs_base;
        ("gs_base", ctypes.c_ulonglong
         ),  # __extension__ unsigned long long int gs_base;
        ("ds",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int ds;
        ("es",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int es;
        ("fs",
         ctypes.c_ulonglong),  # __extension__ unsigned long long int fs;
        ("gs", ctypes.c_ulonglong
         )  # __extension__ unsigned long long int gs;
    ]


# elf_greg_t    = ctypes.c_ulonglong
# ELF_NGREG = ctypes.sizeof(user_regs_struct)/ctypes.sizeof(elf_greg_t)
# elf_gregset_t = elf_greg_t*ELF_NGREG
elf_gregset_t = user_regs_struct


class elf_prstatus(ctypes.Structure):  # struct elf_prstatus
    _fields_ = [
        (
            # Info associated with signal
            # struct elf_siginfo pr_info;
            "pr_info", elf_siginfo
        ),
        (
            # Current signal
            # short int pr_cursig;
            "pr_cursig", ctypes.c_short
        ),
        (
            # Set of pending signals
            # unsigned long int pr_sigpend;
            "pr_sigpend", ctypes.c_ulong
        ),
        (
            # Set of held signals
            # unsigned long int pr_sighold;
            "pr_sighold", ctypes.c_ulong
        ),
        (
            # Process ID
            # __pid_t pr_pid;
            "pr_pid", ctypes.c_int
        ),
        (
            # Parent process ID
            # __pid_t pr_ppid;
            "pr_ppid", ctypes.c_int
        ),
        (
            # Parent group ID
            # __pid_t pr_pgrp;
            "pr_pgrp", ctypes.c_int
        ),
        (
            # Parent session ID
            # __pid_t pr_sid;
            "pr_sid", ctypes.c_int
        ),
        (
            # User time
            # struct timeval pr_utime;
            "pr_utime", timeval
        ),
        (
            # System time
            # struct timeval pr_stime;
            "pr_stime", timeval
        ),
        (
            # Cumulative user time
            # struct timeval pr_cutime;
            "pr_cutime", timeval
        ),
        (
            # Cumulative system time
            # struct timeval pr_cstime;
            "pr_cstime", timeval
        ),
        (
            # GP registers
            # elf_gregset_t pr_reg;
            "pr_reg", elf_gregset_t
        ),
        (
            # True if math copro being used
            # int pr_fpvalid;
            "pr_fpvalid", ctypes.c_int
        )
    ]


# elf_prpsinfo related constants.

# Number of chars for args
# #define ELF_PRARGSZ     (80)
ELF_PRARGSZ = 80


class elf_prpsinfo(ctypes.Structure):  # struct elf_prpsinfo
    _fields_ = [
        (
            # Numeric process state
            # char pr_state;
            "pr_state", ctypes.c_byte
        ),
        (
            # Char for pr_state
            # char pr_sname;
            "pr_sname", ctypes.c_char
        ),
        (
            # Zombie
            # char pr_zomb;
            "pr_zomb", ctypes.c_byte
        ),
        (
            # Nice value
            # char pr_nice;
            "pr_nice", ctypes.c_byte
        ),
        (
            # Flags
            # unsigned long int pr_flag;
            "pr_flag", ctypes.c_ulong
        ),
        (
            # User ID
            # unsigned int pr_uid;
            "pr_uid", ctypes.c_uint
        ),
        (
            # Group ID
            # unsigned int pr_gid;
            "pr_gid", ctypes.c_uint
        ),
        ("pr_pid", ctypes.c_int),
        ("pr_ppid", ctypes.c_int),
        ("pr_pgrp", ctypes.c_int),
        ("pr_sid", ctypes.c_int),
        # /* Lots missing */
        (
            # Filename of executable
            # char pr_fname[16];
            "pr_fname", ctypes.c_char * 16
        ),
        (
            # Initial part of arg list
            # char pr_psargs[ELF_PRARGSZ];
            "pr_psargs", ctypes.c_char * ELF_PRARGSZ
        )
    ]


class user_fpregs_struct(ctypes.Structure):  # struct user_fpregs_struct
    _fields_ = [
        # unsigned short int cwd;
        ("cwd", ctypes.c_ushort),
        # unsigned short int swd;
        ("swd", ctypes.c_ushort),
        # unsigned short int ftw;
        ("ftw", ctypes.c_ushort),
        # unsigned short int fop;
        ("fop", ctypes.c_ushort),
        # __extension__ unsigned long long int rip;
        ("rip", ctypes.c_ulonglong),
        # __extension__ unsigned long long int rdp;
        ("rdp", ctypes.c_ulonglong),
        # unsigned int mxcsr;
        ("mxcsr", ctypes.c_uint),
        # unsigned int mxcr_mask;
        ("mxcr_mask", ctypes.c_uint),
        # unsigned int st_space[32];   /* 8*16 bytes for each FP-reg = 128 bytes */
        ("st_space", ctypes.c_uint * 32),
        # unsigned int xmm_space[64];  /* 16*16 bytes for each XMM-reg = 256 bytes */
        ("xmm_space", ctypes.c_uint * 64),
        # unsigned int padding[24];
        ("padding", ctypes.c_uint * 24),
    ]


elf_fpregset_t = user_fpregs_struct

# siginfo_t related constants.

_SI_MAX_SIZE = 128
_SI_PAD_SIZE = (_SI_MAX_SIZE // ctypes.sizeof(ctypes.c_int)) - 4


# /* kill().  */
class _siginfo_t_U_kill(ctypes.Structure):  # struct
    _fields_ = [
        (
            # Sending process ID
            # __pid_t si_pid;
            "si_pid", ctypes.c_int
        ),
        (
            # Real user ID of sending process
            # __uid_t si_uid;
            "si_uid", ctypes.c_uint
        )
    ]  # } _kill;


# Type for data associated with a signal.
class sigval_t(ctypes.Union):  # typedef union sigval
    _fields_ = [
        ("sival_int", ctypes.c_int),  # int sival_int;
        ("sical_ptr", ctypes.c_void_p),  # void *sival_ptr;
    ]  # } sigval_t;


# /* POSIX.1b timers.  */
class _siginfo_t_U_timer(ctypes.Structure):  # struct
    _fields_ = [
        (
            # Timer ID
            # int si_tid;
            "si_tid", ctypes.c_int
        ),
        (
            # Overrun count
            # int si_overrun;
            "si_overrun", ctypes.c_int
        ),
        (
            # Signal value
            # sigval_t si_sigval;
            "si_sigval", sigval_t
        )
    ]  # } _timer;


# /* POSIX.1b signals.  */
class _siginfo_t_U_rt(ctypes.Structure):  # struct
    _fields_ = [
        (
            # Sending process ID
            # __pid_t si_pid;
            "si_pid", ctypes.c_int
        ),
        (
            # Real user ID of sending process
            # __uid_t si_uid;
            "si_uid", ctypes.c_uint
        ),
        (
            # Signal value
            # sigval_t si_sigval;
            "si_sigval", sigval_t
        )
    ]  # } _rt;


# /* SIGCHLD.  */
class _siginfo_t_U_sigchld(ctypes.Structure):  # struct
    _fields_ = [
        (
            # Which child
            # __pid_t si_pid;
            "si_pid", ctypes.c_int
        ),
        (
            # Real user ID of sending process
            # __uid_t si_uid;
            "si_uid", ctypes.c_uint
        ),
        (
            # Exit value or signal
            # int si_status;
            "si_status", ctypes.c_int
        ),
        (
            # __sigchld_clock_t si_utime;
            "si_utime", ctypes.c_long
        ),
        (
            # __sigchld_clock_t si_stime;
            "si_stime", ctypes.c_long
        )
    ]  # } _sigchld;


# /* SIGILL, SIGFPE, SIGSEGV, SIGBUS.  */
class _siginfo_t_U_sigfault(ctypes.Structure):  # struct
    _fields_ = [
        (
            # Faulting insn/memory ref
            # void *si_addr;
            "si_addr", ctypes.c_void_p
        ),
        (
            # Valid LSB of the reported address
            # short int si_addr_lsb;
            "si_addr_lsb", ctypes.c_short
        )
    ]  # } _sigfault;


# /* SIGPOLL.  */
class _siginfo_t_U_sigpoll(ctypes.Structure):  # struct
    _fields_ = [
        (
            # Band event for SIGPOLL
            # long int si_band;
            "si_band", ctypes.c_long
        ),
        (
            # int si_fd;
            "si_fd", ctypes.c_int
        )
    ]  # } _sigpoll;


# /* SIGSYS.  */
class _siginfo_t_U_sigsys(ctypes.Structure):  # struct
    _fields_ = [
        ("_call_addr", ctypes.c_void_p
         ),  # void *_call_addr;   /* Calling user insn.  */
        (
            "_syscall", ctypes.c_int
        ),  # int _syscall;       /* Triggering system call number.  */
        ("_arch", ctypes.c_uint
         )  # unsigned int _arch; /* AUDIT_ARCH_* of syscall.  */
    ]  # } _sigsys;


class _siginfo_t_U(ctypes.Union):  # union
    _fields_ = [
        ("_pad",
         ctypes.c_int * _SI_PAD_SIZE),  # int _pad[__SI_PAD_SIZE];

        # /* kill().  */
        ("_kill", _siginfo_t_U_kill),  # struct

        # __pid_t si_pid;     /* Sending process ID.  */
        # __uid_t si_uid;     /* Real user ID of sending process.  */
        # } _kill;

        # /* POSIX.1b timers.  */
        ("_timer", _siginfo_t_U_timer),  # struct

        # int si_tid;         /* Timer ID.  */
        # int si_overrun;     /* Overrun count.  */
        # sigval_t si_sigval; /* Signal value.  */
        # } _timer;

        # /* POSIX.1b signals.  */
        ("_rt", _siginfo_t_U_rt),  # struct

        # __pid_t si_pid;     /* Sending process ID.  */
        # __uid_t si_uid;     /* Real user ID of sending process.  */
        # sigval_t si_sigval; /* Signal value.  */
        # } _rt;

        # /* SIGCHLD.  */
        ("_sigchld", _siginfo_t_U_sigchld),  # struct

        # __pid_t si_pid;     /* Which child.  */
        # __uid_t si_uid;     /* Real user ID of sending process.  */
        # int si_status;      /* Exit value or signal.  */
        # __sigchld_clock_t si_utime;
        # __sigchld_clock_t si_stime;
        # } _sigchld;

        # /* SIGILL, SIGFPE, SIGSEGV, SIGBUS.  */
        ("_sigfault", _siginfo_t_U_sigfault),  # struct

        # void *si_addr;      /* Faulting insn/memory ref.  */
        # short int si_addr_lsb;      /* Valid LSB of the reported address.  */
        # } _sigfault;

        # /* SIGPOLL.  */
        ("_sigpoll", _siginfo_t_U_sigpoll),  # struct

        # long int si_band;   /* Band event for SIGPOLL.  */
        # int si_fd;
        # } _sigpoll;

        # /* SIGSYS.  */
        ("_sigsys", _siginfo_t_U_sigpoll)  # struct

        # void *_call_addr;   /* Calling user insn.  */
        # int _syscall;       /* Triggering system call number.  */
        # unsigned int _arch; /* AUDIT_ARCH_* of syscall.  */
        # } _sigsys;
    ]  # } _sifields;


class siginfo_t(ctypes.Structure):  # typedef struct
    _fields_ = [
        (
            # Signal number
            # int si_signo;
            "si_signo", ctypes.c_int
        ),
        (
            # If non-zero, an errno value associated with
            # int si_errno;
            "si_errno", ctypes.c_int
        ),
        (
            # Signal code - this signal, as defined in <errno.h>
            # int si_code;
            "si_code", ctypes.c_int
        ),
        (
            # Union
            "_sifields", _siginfo_t_U
        )

        # int _pad[__SI_PAD_SIZE];
        #
        # /* kill().  */
        # struct

        # __pid_t si_pid;     /* Sending process ID.  */
        # __uid_t si_uid;     /* Real user ID of sending process.  */
        # } _kill;
        #
        # /* POSIX.1b timers.  */
        # struct

        # int si_tid;         /* Timer ID.  */
        # int si_overrun;     /* Overrun count.  */
        # sigval_t si_sigval; /* Signal value.  */
        # } _timer;
        #
        # /* POSIX.1b signals.  */
        # struct

        # __pid_t si_pid;     /* Sending process ID.  */
        # __uid_t si_uid;     /* Real user ID of sending process.  */
        # sigval_t si_sigval; /* Signal value.  */
        # } _rt;
        #
        # /* SIGCHLD.  */
        # struct

        # __pid_t si_pid;     /* Which child.  */
        # __uid_t si_uid;     /* Real user ID of sending process.  */
        # int si_status;      /* Exit value or signal.  */
        # __sigchld_clock_t si_utime;
        # __sigchld_clock_t si_stime;
        # } _sigchld;
        #
        # /* SIGILL, SIGFPE, SIGSEGV, SIGBUS.  */
        # struct

        # void *si_addr;      /* Faulting insn/memory ref.  */
        # short int si_addr_lsb;      /* Valid LSB of the reported address.  */
        # } _sigfault;
        #
        # /* SIGPOLL.  */
        # struct

        # long int si_band;   /* Band event for SIGPOLL.  */
        # int si_fd;
        # } _sigpoll;
        #
        # /* SIGSYS.  */
        # struct

        # void *_call_addr;   /* Calling user insn.  */
        # int _syscall;       /* Triggering system call number.  */
        # unsigned int _arch; /* AUDIT_ARCH_* of syscall.  */
        # } _sigsys;
        # } _sifields;
    ]  # } siginfo_t __SI_ALIGNMENT;


# xsave related.


class ymmh_struct(ctypes.Structure):  # struct ymmh_struct {
    _fields_ = [
        # u32 ymmh_space[64];
        ("ymmh_space", 64 * ctypes.c_uint)
    ]  # } __packed;


class xsave_hdr_struct(ctypes.Structure):  # struct xsave_hdr_struct {
    _fields_ = [
        # u64 xstate_bv;
        ("xstate_bv", ctypes.c_ulonglong),
        # u64 reserved1[2];
        ("reserved1", ctypes.c_ulonglong * 2),
        # u64 reserved2[5];
        ("reserved2", ctypes.c_ulonglong * 5)
    ]  # } __packed;


class i387_fxsave_struct(ctypes.Structure):  # struct i387_fxsave_struct {
    _fields_ = [
        (
            # Control Word
            # u16 cwd;
            "cwd", ctypes.c_ushort
        ),
        (
            # Status Word
            # u16 swd;
            "swd", ctypes.c_ushort
        ),
        (
            # Tag Word
            # u16 twd;
            "twd", ctypes.c_ushort
        ),
        (
            # Last Instruction Opcode
            # u16 fop;
            "fop", ctypes.c_ushort
        ),
        # union {
        # struct {
        (
            # Instruction Pointer
            # u64 rip;
            "rip", ctypes.c_ulonglong
        ),
        (
            # Data Pointer
            # u64 rdp;
            "rdp", ctypes.c_ulonglong
        ),

        # struct {
        # u32             fip; /* FPU IP Offset                   */
        # u32             fcs; /* FPU IP Selector                 */
        # u32             foo; /* FPU Operand Offset              */
        # u32             fos; /* FPU Operand Selector            */

        (
            # MXCSR Register State
            # u32 mxcsr;
            "mxcsr", ctypes.c_uint
        ),
        (
            # MXCSR Mask
            # u32 mxcsr_mask;
            "mxcsr_mask", ctypes.c_uint
        ),
        # 8*16 bytes for each FP-reg = 128 bytes
        (
            # u32 st_space[32];
            "st_space", ctypes.c_uint * 32
        ),
        # 16*16 bytes for each XMM-reg = 256 bytes
        (
            # u32 xmm_space[64];
            "xmm_space", ctypes.c_uint * 64
        ),
        (
            # u32 padding[12];
            "padding", ctypes.c_uint * 12
        ),
        # union {
        (
            # u32 padding1[12];
            "padding1", ctypes.c_uint * 12
        )
        # u32                     sw_reserved[12];
    ]  # } __aligned(16);


class elf_xsave_struct(ctypes.Structure):  # struct xsave_struct {
    _fields_ = [
        # struct i387_fxsave_struct i387;
        ("i387", i387_fxsave_struct),
        # struct xsave_hdr_struct xsave_hdr;
        ("xsave_hdr", xsave_hdr_struct),
        # struct ymmh_struct              ymmh;
        ("ymmh", ymmh_struct)
    ]  # } __aligned(FP_MIN_ALIGN_BYTES) __packed;
